@@ -11,9 +11,12 @@ from os import R_OK, access
 try:
     from os import strerror
 except ImportError:
-    strerror = lambda: 'strerror not supported'
-from os.path import exists
+    strerror = lambda x: 'strerror not supported'
+from os.path import exists, getsize
 from itertools import izip
+import logging
+import textwrap
+import sys
 
 def true_long_type():
     """
@@ -82,7 +85,8 @@ def create_twobyte_table():
 BYTE_TABLE = create_byte_table()
 TWOBYTE_TABLE = create_twobyte_table()
 
-def longs_to_char_array(longs, first_base_offset, last_base_offset, array_size):
+def longs_to_char_array(longs, first_base_offset, last_base_offset, array_size,
+                        more_bytes=None):
     """
     takes in a iterable of longs and converts them to bases in a char array
     returns a ctypes string buffer
@@ -99,14 +103,26 @@ def longs_to_char_array(longs, first_base_offset, last_base_offset, array_size):
     i = 16 - first_base_offset
     if array_size < i: i = array_size
     dna[0:i] = array('c', first_block[first_base_offset:first_base_offset + i])
-    if longs_len == 1: return dna
-    # middle blocks (implicitly skipped if they don't exist)
-    for byte in bytes[4:-4]:
-        dna[i:i + 4] = array('c', BYTE_TABLE[byte])
-        i += 4
-    # last block
-    last_block = array('c', ''.join([''.join(BYTE_TABLE[bytes[x]]) for x in range(-4,0)]))
-    dna[i:i + last_base_offset] = last_block[0:last_base_offset]
+    if longs_len > 1:
+        # middle blocks (implicitly skipped if they don't exist)
+        for byte in bytes[4:-4]:
+            dna[i:i + 4] = array('c', BYTE_TABLE[byte])
+            i += 4
+        # last block
+        last_block = array('c', ''.join([''.join(BYTE_TABLE[bytes[x]]) for x in range(-4,0)]))
+        dna[i:i + last_base_offset] = last_block[0:last_base_offset]
+        i += 16
+    if more_bytes is not None:
+        bytes = array('B')
+        bytes.fromstring(more_bytes)
+        j = i
+        for byte in bytes:
+            j = i + 4
+            if j > array_size:
+                dna[i:array_size] = array('c', BYTE_TABLE[byte])[0:(array_size - i)]
+                break
+            dna[i:i + 4] = array('c', BYTE_TABLE[byte])
+            i += 4
     return dna
 
 class TwoBitFile(dict):
@@ -133,17 +149,19 @@ See TwoBitSequence for more info
     """
     
     def __init__(self, foo):
-        super(dict, self).__init__(self)
+        super(TwoBitFile, self).__init__()
         if not exists(foo):
             raise IOError(ENOENT, strerror(ENOENT), foo)
         if not access(foo, R_OK):
             raise IOError(EACCES, strerror(EACCES), foo)
         self._filename = foo
+        self._file_size = getsize(foo)
         self._file_handle = open(foo, 'rb')
         self._load_header()
         self._load_index()
         for name, offset in self._offset_dict.iteritems():
             self[name] = TwoBitSequence(self._file_handle, offset,
+                                        self._file_size,
                                         self._byteswapped)
         return        
         
@@ -158,12 +176,13 @@ See TwoBitSequence for more info
         if not signature == 0x1A412743:
             byteswapped = True
             header.byteswap()
-            (signature, version, sequence_count, reserved) = header
-            if not signature == 0x1A412743:
-                raise TwoBitFileError('Signature in header should be 0x1A412743.')
+            (signature2, version, sequence_count, reserved) = header
+            if not signature2 == 0x1A412743:
+                raise TwoBitFileError('Signature in header should be 0x1A412743'
+                                    + ', instead found 0x%X' % signature)
         if not version == 0: 
             raise TwoBitFileError('File version in header should be 0.')
-        if not reserved == 0: 
+        if not reserved == 0:
             raise TwoBitFileError('Reserved field in header should be 0.')
         self._byteswapped = byteswapped
         self._sequence_count = sequence_count
@@ -174,14 +193,13 @@ See TwoBitSequence for more info
         remaining = self._sequence_count
         sequence_offsets = []
         file_handle.seek(16)
-        while True:
-            if remaining == 0: break
+        while remaining > 0:
             name_size = array('B')
             name_size.fromfile(file_handle, 1)
             if byteswapped: name_size.byteswap()
             name = array('c')
-            if byteswapped: name.byteswap()
             name.fromfile(file_handle, name_size[0])
+            if byteswapped: name.byteswap()
             offset = array(LONG)
             offset.fromfile(file_handle, 1)
             if byteswapped: offset.byteswap()
@@ -231,7 +249,8 @@ x = TwoBitFile('my.2bit')
 d = x.dict()
 for k,v in d.iteritems(): d[k] = str(v)
     """
-    def __init__(self, file_handle, offset, byteswapped=False):
+    def __init__(self, file_handle, offset, file_size, byteswapped=False):
+        self._file_size = file_size
         self._file_handle = file_handle
         self._original_offset = offset
         self._byteswapped = byteswapped
@@ -240,8 +259,9 @@ for k,v in d.iteritems(): d[k] = str(v)
         header.fromfile(file_handle, 2)
         if byteswapped: header.byteswap()
         dna_size, n_block_count = header
-        self._dna_size = dna_size
-        self._packed_dna_size = (dna_size + 15) / 16 # this is 32-bit fragments
+        self._dna_size = dna_size # number of characters, 2 bits each
+        self._n_bytes = (dna_size + 3) / 4 # number of bytes
+        self._packed_dna_size = (dna_size + 15) / 16 # number of 32-bit fragments
         n_block_starts = array(LONG)
         n_block_sizes = array(LONG)
         n_block_starts.fromfile(file_handle, n_block_count)
@@ -268,26 +288,26 @@ for k,v in d.iteritems(): d[k] = str(v)
     def __len__(self):
         return self._dna_size
 
-    def __getslice__(self, min, max=None):
-        return self.get_slice(min, max)
+    def __getslice__(self, min_, max_=None):
+        return self.get_slice(min_, max_)
 
-    def get_slice(self, min, max=None):
+    def get_slice(self, min_, max_=None):
         """
         get_slice returns only a sub-sequence
         """
         # handle negative coordinates
         dna_size = self._dna_size
-        if max < 0:
-            if max < -dna_size: raise IndexError('index out of range')
-            max = dna_size + 1 + max
-        if min < 0:
-            if max < -dna_size: raise IndexError('index out of range')
-            min = dna_size + 1 + min
+        if max_ is not None and max_ < 0:
+            if max_ < -dna_size: raise IndexError('index out of range')
+            max_ = dna_size + 1 + max_
+        if min_ < 0:
+            if max_ < -dna_size: raise IndexError('index out of range')
+            min_ = dna_size + 1 + min_
         # make sure there's a proper range
-        if min > max and max is not None: return ''
-        if max == 0: return ''
+        if max_ is not None and min_ > max_: return ''
+        if max_ == 0 or max_ == min_: return ''
         # load all the data
-        if max > dna_size: max = dna_size
+        if max_ is None or max_ > dna_size: max_ = dna_size
         file_handle = self._file_handle
         byteswapped = self._byteswapped
         n_block_starts = self._n_block_starts
@@ -296,60 +316,74 @@ for k,v in d.iteritems(): d[k] = str(v)
         mask_block_sizes = self._mask_block_sizes
         offset = self._offset
         packed_dna_size = self._packed_dna_size
+        n_bytes = self._n_bytes
 
         # region_size is how many bases the region is       
-        if max is None: region_size = dna_size - min
-        else: region_size = max - min
+        if max_ is None: region_size = dna_size - min_
+        else: region_size = max_ - min_
         
         # start_block, end_block are the first/last 32-bit blocks we need
-        # note: end_block is not read
         # blocks start at 0
-        start_block = min / 16
-        end_block = (max - 1) / 16
-        # don't read past seq end
-        if end_block >= packed_dna_size: end_block = packed_dna_size - 1
-        # +1 we still need to read block
-        blocks_to_read = end_block - start_block + 1
-        
+        start_block = min_ / 16
         # jump directly to desired file location
-        local_offset = offset + start_block * 4
+        local_offset = offset + (start_block * 4)
+        end_block = (max_ - 1 + 16) / 16
+        # don't read past seq end
+        
         file_handle.seek(local_offset)
         
         # note we won't actually read the last base
         # this is a python slice first_base_offset:16*blocks+last_base_offset
-        first_base_offset = min % 16
-        last_base_offset = max % 16
+        first_base_offset = min_ % 16
+        last_base_offset = max_ % 16
+        if last_base_offset == 0: last_base_offset = 16
+        # +1 we still need to read end_block maybe
+        blocks_to_read = end_block - start_block
+        if (blocks_to_read + start_block) > packed_dna_size:
+            blocks_to_read = packed_dna_size - start_block
         
         fourbyte_dna = array(LONG)
-        fourbyte_dna.fromfile(file_handle, blocks_to_read)
+        remainder_seq = None
+        if (blocks_to_read * 4 + local_offset) > self._file_size:
+            fourbyte_dna.fromfile(file_handle, blocks_to_read - 1)
+            morebytes = file_handle.read() # read the remaining characters
+#            if byteswapped:
+#                morebytes = ''.join(reversed(morebytes))
+        else:
+            fourbyte_dna.fromfile(file_handle, blocks_to_read)
+            morebytes = None
         if byteswapped: fourbyte_dna.byteswap()
         string_as_array = longs_to_char_array(fourbyte_dna, first_base_offset,
-                                              last_base_offset, region_size)
+                                              last_base_offset, region_size,
+                                              more_bytes=morebytes)
         for start, size in izip(n_block_starts, n_block_sizes):
             end = start + size
-            if end <= min: continue
-            if start > max: break
-            if start < min: start = min
-            if end > max: end = max 
-            start -= min
-            end -= min
-            string_as_array[start:end] = 'N'*(end-start)
+            if end <= min_: continue
+            if start > max_: break
+            if start < min_: start = min_
+            if end > max_: end = max_ 
+            start -= min_
+            end -= min_
+            # this should actually be decoded, 00=N, 01=n
+            string_as_array[start:end] = array('c', 'N'*(end-start))
         lower = str.lower
         first_masked_region = max(0,
-                                  bisect_right(mask_block_starts, min) - 1)
-        last_masked_region = min(len(masked_block_starts),
-                                 1 + bisect_right(mask_block_starts, max,
+                                  bisect_right(mask_block_starts, min_) - 1)
+        last_masked_region = min(len(mask_block_starts),
+                                 1 + bisect_right(mask_block_starts, max_,
                                                   lo=first_masked_region))
         for start, size in izip(mask_block_starts[first_masked_region:last_masked_region],
                                 mask_block_sizes[first_masked_region:last_masked_region]):
             end = start + size
-            if end <= min: continue
-            if start > max: break
-            if start < min: start = min
-            if end > max: end = max 
-            start -= min
-            end -= min
+            if end <= min_: continue
+            if start > max_: break
+            if start < min_: start = min_
+            if end > max_: end = max_ 
+            start -= min_
+            end -= min_
             string_as_array[start:end] = array('c', lower(string_as_array[start:end].tostring()))
+        if not len(string_as_array) == max_ - min_:
+            raise RuntimeError, "Sequence was the wrong size"
         return string_as_array.tostring()
 
     def __str__(self):
@@ -419,4 +453,100 @@ represented as 00011011. The packedDna field will be padded with 0 bits as \
 necessary so that it takes an even multiple of 32 bit in the file, as this \
 improves i/o performance on some machines.
 .nib files
-"""
\ No newline at end of file
+"""
+
+def cmdline_reader():
+    """
+    cmdline_reader allows twobitreader module to be executed as a script
+    accepts only one argument -- the .2bit filename
+    reads input (BED format) from stdin
+    writes output (FASTA format) to stdout
+    writes errors/warning to stderr
+    
+    Regions should be given in BED format on stdin
+    chrom    start(0-based)    end(0-based, not included)
+    
+    To use a BED file of regions, do
+    python -m twobitreader example.2bit < example.bed
+    
+    Non-regions will be skipped and warnings will be issued to logging
+    (logging output to stderr by default)
+    """
+    # if no argument provided, print docstring
+    argv = sys.argv
+    if len(argv) == 1:
+        print argv[0] + ":"
+        print cmdline_reader.__doc__
+        sys.exit()
+        return
+    # if user is trying to get help, print docstring
+    elif len(argv) == 2 and argv[1] in ['--help', '-h', '-help']:
+        print argv[0] + ":"
+        print cmdline_reader.__doc__
+        sys.exit()
+        return
+    # if user specified multiple files, exit with error
+    elif len(argv) > 2:
+        sys.exit("Too many files specified")
+        return
+    # otherwise proceed with opening the .2bit file
+    twobit_file = TwoBitFile(argv[1])
+    # print error/warning messages as we go
+    warning_msg = 'Invalid %s at line %d\n\t"%s"'
+    twobit_reader(twobit_file, input_stream=sys.stdin)
+
+def twobit_reader(twobit_file, input_stream=None, write=None):
+    """
+    twobit_reader takes a twobit_file (of class TwoBitFile)
+    and an "input_stream" which can be any iterable (incl. file-like objects)
+    writes output (FASTA format) using write (default is to print, if write=None)
+    logs errors/warning to stderr
+    
+    Regions should be given in BED format on stdin
+    chrom    start(0-based)    end(0-based, not included)
+    
+    To use a BED file of regions, do
+    python -m twobitreader example.2bit < example.bed
+    
+    Non-regions will be skipped and warnings will be issued to logging
+    (logging output to stderr by default)
+    """
+    warning_msg = 'Invalid %s at line %d\n\t"%s"'
+    if input_stream is None: return
+    for i, line in enumerate((line.rstrip('\n\r') for line in input_stream)):
+        fields = line.split()
+        if not len(fields) >= 3:
+            logging.warn(warning_msg, 'start', i, line)
+            continue
+        chrom = fields[0]
+        if not twobit_file.has_key(chrom):
+            logging.warn(warning_msg, 'chrom', i, line)
+            continue
+        try: start = long(fields[1])
+        except ValueError:
+            logging.warn(warning_msg, 'start', i, line)
+        if start < 0:
+            logging.warn(warning_msg, 'start', i, line)
+            logging.warn('Using 0 as start instead for line %d', i)
+            start = 0
+        try: end = long(fields[2])
+        except ValueError:
+            logging.warn(warning_msg, 'end', i, line)
+            continue
+        chrom_len = len(twobit_file[chrom])
+        if end > len(twobit_file[chrom]):
+            logging.warn('At line %d, end is greater than chrom length %d\n%s',
+                         i, chrom_len, line)
+            logging.warn('Sequence will be truncated at chrom' +
+                         'length for line %d', i)
+            end = chrom_len
+        seq = twobit_file[chrom][start:end]
+        if write is not None:
+            write(">%s:%d-%d" % (chrom, start, end))
+            write(textwrap.fill(seq, 60))
+        else:
+            print ">%s:%d-%d" % (chrom, start, end)
+            print textwrap.fill(seq, 60)
+    return
+
+if __name__ == '__main__': cmdline_reader()
\ No newline at end of file
